######################################### IMPORTS #############################################
## ALIASED IMPORTS
import IPython.display as ip_display
import numpy as np
import pandas as pd

## NORMAL IMPORTS
import __main__
import datetime 
import time 
import itables
import logging
import importlib
import sys
import os
import gc
import pickle
import functools
import types 

## FUNCTION IMPORTS 
from collections import defaultdict

## LOCAL FUNCTION IMPORTS
from krezi.multiprocessing_util.mp_util import run_in_parallel
from krezi.logging_util.file_logger import Logger

############################# INITIALIZING USEFUL GLOBAL VARIABLES #################################
home_dir = os.environ.get('HOME_DIR') 
notebooks_dir = os.environ.get('NOTEBOOKS_DIR')
data_dir = os.environ.get('DATA_DIR') 
log_dir = os.environ.get('LOG_DIR') 
tmp_dir = os.environ.get('TMP_DIR')

####################################### SETTING LOGGER ############################################
logger = Logger('krezi_def')
log = logger.get_logger() # You can change the date format here

log_debug = log.debug
log_info = log.info
log_warning = log.warning
log_error = log.error
log_critical = log.critical

log_info("Logger initialized WITHOUT file handler")

def init_logger(filepath):
    global log
    log = logger.add_file_handler(filepath)
    log_info(f"File Handler added. Location set to {filepath!r}")

###################################### COMMON FUNCTIONS ##########################################
def fit_to_screen():
    """
    A function to fit the display to the screen by modifying the HTML style. Useful with Jupyter Notebooks.
    """
    ip_display.display(ip_display.HTML("<style>.container { width:100% !important; }</style>"))
fit_to_screen()

## RELATED TO MODULES AND OBJECTS
def imported_modules():
    """
    Generates a generator that yields the names and values of imported modules.

    This function iterates over the global namespace and checks if each value is an instance of the `types.ModuleType` class. If it is, the function yields a tuple containing the module's name and the corresponding variable name.

    Yields:
        tuple: A tuple containing the module name (str) and the variable name (str) of an imported module.

    Example:
        >>> for module_name, var_name in imported_modules():
        ...     print(module_name, var_name)
        module1 var1
        module2 var2
    """
    for name, val in globals().items():  # Everytime Globals is called, dict size will change as every run adds a new cell run info like _i120 and _120
        if isinstance(val, types.ModuleType):
            yield val.__name__, name

def display_imported_modules():
    """
    Display the imported modules and their aliases.

    This function iterates over the imported modules and their aliases using the `imported_modules` function. It prints the module name followed by its alias, with a fixed width of 50 characters for the module name and 5 spaces for the alias.

    Parameters:
        None

    Returns:
        None
    """
    for module, alias in imported_modules(): 
        print(f"{module:<50} as {' '*5}{alias}")

def reload_modules(modules:list) -> None:
    """
    Reloads the specified modules.

    Args:
        modules (list): A list of modules to reload.

    Returns:
        None
    """
    for module in modules:
        importlib.reload(module)

def reload_krezi():
    """
    Reloads the 'krezi' module.

    This function reloads the 'krezi' module by calling the 'reload_modules' function with the list containing the 'krezi' module as its argument.

    Parameters:
        None

    Returns:
        None
    """
    reload_modules(['krezi'])

import ctypes
def get_object_at_address(address):
    '''
    address=0x7822e7c7bac0
    print(f"Value from address {address} is {ctypes.cast(address, ctypes.py_object).value}")
    '''
    obj = ctypes.cast(address, ctypes.py_object).value
    return obj

def elapsed_time(start_tm):
    """
    Calculate the elapsed time from a given start time.

    Args:
        start_tm (float): The start time in seconds since the epoch.

    Returns:
        str: The elapsed time in seconds if it is less than 120 seconds, otherwise in minutes.
    """
    et = time.time() - start_tm
    if et < 120 : 
        return f"{et:.1f} seconds"
    else : 
        return f"{et/60:.1f} minutes"

# ^Decorator
def func_et(func):
    """
    Decorator function that logs the start and end of a function execution, as well as the time taken to execute it.

    Parameters:
        func (function): The function to be wrapped.

    Returns:
        function: The wrapped function.

    Example:
        @func_et
        def my_function():
            # Function code here
            pass

        # Output:
        # log_prefix :: my_function :: Function Execution Started
        # log_prefix :: my_function :: Function Execution Done :: Time Taken -> 0.123 seconds
    """
    @functools.wraps
    def func_wrapper(*args, **kwargs):
        log_info(f"{func.__name__} :: Function Execution Started")
        start_tm = time.time()
        res = func(*args, **kwargs)
        time_taken = elapsed_time(start_tm)
        log_info(f"{func.__name__} :: Function Execution Done :: Time Taken -> {time_taken}")
        if res: return res
    return func_wrapper

def flatten_list_of_lists(t):
    """
    Flattens a list of lists into a single list.

    Args:
        t (list): A list of lists to be flattened.

    ALternative:
        Use itertool's chain method.

    Returns:
        list: A flattened list containing all the elements from the input list of lists.
    """
    return [item for sublist in t for item in sublist]

from tqdm.notebook import tqdm_notebook
def tqdm_custom(iterable,**add_params):
    """
    Creates a custom tqdm progress bar for iterating over an iterable.

    Args:
        iterable (iterable): The iterable to iterate over.
        **add_params (dict): Additional parameters to pass to tqdm_notebook.

    Returns:
        tqdm.notebook.tqdm_notebook: A tqdm progress bar for iterating over the iterable.

    Note:
        The custom bar format includes the description, percentage, bar, current and total item counts, elapsed time, and remaining time.
    """
    bar_format = "{desc}: {percentage:.3f}%|{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining} {rate_inv_fmt}"
    return tqdm_notebook(iterable,bar_format = bar_format,**add_params)

## RELATED TO SUBPROCESSING MODULE
# command = f"aws s3 ls {s3_dir} --human-readable --summarize | sort -k 2 -r | awk '{{print $5}}'"
import subprocess
def run_command(command, return_output = False):
    with subprocess.Popen(command, stdout = subprocess.PIPE, stderr = subprocess.PIPE, shell=True, universal_newlines=True) as p:
        log_info(f"Running Command (first 200 characters) :: {command[:200]}")
        output = p.stdout.readlines()
        log_info("output :: ", ''.join(output))
        log_info("error :: ", p.stderr.read())
    log_info("return code :: ", p.returncode)
    if return_output:
            res = set(output)
            res.discard('\n')
            return list(res)
cmd = run_command

def run_command_return_output(command):
    return run_command(command, return_output = True)

## RELATED TO IMPORT EXPORT
def to_pickle(object, path):
    with open(path, 'wb') as f:
        pickle.dump(object, f)

def read_pickle(path):
    with open(path, 'rb') as f:
        object = pickle.load(f)
    return object

def read_parquet_from_s3(s3_path, FILENAME):
    this_file_name = FILENAME

    filename = this_file_name if FILENAME.endswith(".parquet") else f'{this_file_name}.parquet'
    s3_dir = 's3://analytics.faasos.io'
    s3_path = f'{s3_dir}/{s3_path}'
    s3_path = s3_path if s3_path.endswith("/") else s3_path+"/"

    cmd(f"aws s3 cp {s3_path}{filename} /tmp/")

    return pd.read_parquet(f'/tmp/{filename}')

def to_parquet_to_s3(df, s3_path, FILENAME):
    this_file_name = FILENAME

    filename = this_file_name if FILENAME.endswith(".parquet") else f'{this_file_name}.parquet'
    df.to_parquet(f'/tmp/{filename}', index=False)
    
    s3_dir = 's3://analytics.faasos.io'
    s3_path = f'{s3_dir}/{s3_path}'
    s3_path = s3_path if s3_path.endswith("/") else s3_path+"/"

    cmd(f"aws s3 cp /tmp/{filename} {s3_path}")

## RELATED TO MATHS
def custom_round_nearest_p5(number):
    """
    Rounds a number to the nearest multiple of 0.5.

    Parameters:
        number (float): The number to be rounded.

    Returns:
        float: The rounded number.

    Example:
        >>> custom_round_nearest_p5(2.3)
        2.5
        >>> custom_round_nearest_p5(3.7)
        4.0
        >>> custom_round_nearest_p5(0.0)
        0.0
    """
    x = number%1
    if x==0:
        return number
    elif x<=0.5:
        return int(number)+0.5
    elif x>0.5:
        return int(number)+1.0

def custom_mode(arr):
    freq = defaultdict(int)
    for i in arr:
        freq[i] += 1
    max_freq = max(freq.values())
    max_at = [k for k,v in freq.items() if v == max_freq]
    if len(max_at) > 1:
        return -1
    else:
        return max_at[0]

def custom_mode_return_all(arr):
    """
    Returns a list of all elements in the input array that have the highest frequency.
    
    Parameters:
        arr (Iterable): An iterable of elements.
        
    Returns:
        List[Any]: A list of elements that have the highest frequency in the input array.
    """
    freq = defaultdict(int)
    for i in arr:
        freq[i] += 1
    max_freq = max(freq.values())
    max_at = [k for k,v in freq.items() if v == max_freq]
    return max_at

## RELATED TO PANDAS
## Printing Dataframes as DataTables 
# https://github.com/mwouts/itables/blob/61c1c916175a77f27623eb93fa4ecf42b9a7b7b4/itables/options.py
# for options

pd.DataFrame.data_table = itables.show
pd.DataFrame.dT = pd.DataFrame.data_table

STARTING_THRESHOLD = 0
def accuracy_calculator_xgb_reg(y_actual, y_pred, apply_round = False, round_engine = custom_round_nearest_p5, thresh = STARTING_THRESHOLD):
    """
    Calculates the accuracy of a regression model using XGBoost.
    
    Args:
        y_actual (array-like): The actual values.
        y_pred (array-like): The predicted values.
        apply_round (bool, optional): Whether to apply rounding to the actual and predicted values. Defaults to False.
        round_engine (function, optional): The rounding function to apply. Defaults to custom_round_nearest_p5.
        thresh (float, optional): The threshold value for calculating the accuracy. Defaults to STARTING_THRESHOLD.
        
    Returns:
        dict: A dictionary containing the accuracy metrics:
            - "ACCURACY" (float): The accuracy in percentage.
            - "MAPE" (float): The mean absolute percentage error.
            - "MAE" (float): The mean absolute error.
            - "ACC_10%" (float): The accuracy at 10% tolerance level.
            - "ACC_15%" (float): The accuracy at 15% tolerance level.
            - "ACC_20%" (float): The accuracy at 20% tolerance level.
            - "ACC_30%" (float): The accuracy at 30% tolerance level.
            - "over_predicted%" (float): The percentage of over-predicted values.
            - "under_predicted%" (float): The percentage of under-predicted values.
    """
    
    if apply_round:
        y_actual = pd.Series(y_actual).apply(round_engine).reset_index(drop = True)
        y_pred = pd.Series(y_pred).apply(round_engine).reset_index(drop = True)
    
    non_zero_index = y_actual>0
    
    y_actual = y_actual[non_zero_index]
    y_pred = y_pred[non_zero_index]
    
    if len(y_actual)!=len(y_pred):
        raise ValueError(f"Length of y_actual and y_pred must be same. Please check and try again. Found {len(y_actual)} & {len(y_pred)}")
    
    acc_mape = 100 - np.mean(np.abs((y_actual-y_pred)/y_actual*100))
    acc_mae = np.mean(np.abs(y_actual-y_pred))
    
    if len(y_actual) > 0:
        acc_10_perc = sum(((y_pred >= y_actual * 0.9) & (y_pred <= y_actual * 1.1)) | (np.abs(y_pred-y_actual) <=thresh)) / len(y_actual)*100
        acc_15_perc = sum(((y_pred >= y_actual * 0.85) & (y_pred <= y_actual * 1.15)) | (np.abs(y_pred-y_actual) <=thresh)) / len(y_actual)*100
        acc_20_perc = sum(((y_pred >= y_actual * 0.8) & (y_pred <= y_actual * 1.2)) | (np.abs(y_pred-y_actual) <=thresh)) / len(y_actual)*100
        acc_30_perc = sum(((y_pred >= y_actual * 0.7) & (y_pred <= y_actual * 1.3)) | (np.abs(y_pred-y_actual) <=thresh)) / len(y_actual)*100
        perc_op = sum((y_pred >= y_actual ))/len(y_actual)*100
        perc_up = sum((y_pred < y_actual ))/len(y_actual)*100
    else:
        acc_10_perc=acc_15_perc=acc_20_perc=perc_op=perc_up = np.nan
    
    return {"ACCURACY" : round(acc_mape,2),
            "MAPE" :round(100-acc_mape,2),
            "MAE" : round(acc_mae,2),
            "ACC_10%" : round(acc_10_perc,2),
            "ACC_15%" : round(acc_15_perc,2),
            "ACC_20%" : round(acc_20_perc,2),
            "ACC_30%" : round(acc_30_perc,2),
            "over_predicted%" : round(perc_op,2), 
            "under_predicted%" : round(perc_up,2)}

def to_map(self, key_col, value_col):
    """
    Generates a dictionary mapping values from the specified columns in the DataFrame.

    Parameters:
        key_col (str): The name of the column to be used as keys in the dictionary.
        value_col (str): The name of the column to be used as values in the dictionary.

    Returns:
        dict: A dictionary mapping values from the key_col to values from the value_col.
    """
    return self.set_index(key_col)[value_col].to_dict()
pd.DataFrame.to_map = to_map

def create_date_features(df, date_col, holidays = None):
    """
    Creates date features for a given DataFrame.

    Parameters:
        df (pandas.DataFrame): The DataFrame containing the date column.
        date_col (str): The name of the date column.
        holidays (pandas.DataFrame, optional): A DataFrame containing holiday information. Defaults to None.

    Returns:
        pandas.DataFrame: The DataFrame with additional date features.

    Raises:
        None
    """
    # * Date Features
    df[date_col] = pd.to_datetime(df[date_col])
    df['date'] = df[date_col]
    df['month'] = df.date.dt.month
    df['year'] = df.date.dt.year
    df['day'] = df.date.dt.day
    df['dayofyear'] = df.date.dt.dayofyear
    df['dayofweek'] = df.date.dt.dayofweek
    df['weekofyear'] = df.date.dt.weekofyear
    df['is_weekend'] = [1 if x.dayofweek in [5,6] else 0 for x in df['date']]
    df['daysinmonth'] = df.date.dt.daysinmonth
    
    # * checks for holidays
    if holidays:
        public_holiday_list = set(holidays[holidays['holiday']=='public'].ds)
        not_pulic_holiday_list = set(holidays[holidays['holiday']!='public'].ds)
    
        df['is_public_holiday'] = df["date"].apply(lambda x: 1 if x in public_holiday_list else 0)
        df['is_not_public_holiday'] = df["date"].apply(lambda x: 1 if x in not_pulic_holiday_list else 0)
    
    # * start and end features
    df['is_month_start'] = df.date.dt.is_month_start.astype(int)
    df['is_month_end'] = df.date.dt.is_month_end.astype(int)
    df['is_quarter_start'] = df.date.dt.is_quarter_start.astype(int)
    df['is_quarter_end'] = df.date.dt.is_quarter_end.astype(int)
    df['is_year_start'] = df.date.dt.is_year_start.astype(int)
    df['is_year_end'] = df.date.dt.is_year_end.astype(int)
    
    # * Season
    # TODO: As seasons dont necessarily have a hierarchy, consider adding is_<_season_> features
    # 0: Winter - 1: Spring - 2: Summer - 3: Fall
    df['season'] = 1
    df.loc[df['month'].isin([12, 1, 2]), 'season'] = 0
    df.loc[df['month'].isin([6, 7, 8]), 'season'] = 2
    df.loc[df['month'].isin([9, 10, 11]), 'season'] = 3
    
    # * Date Year Combination
    df['day^year'] = np.log((np.log(df['dayofyear'] + 1)) ** (df['year'] - 2000))
    
    return df

pd.DataFrame.create_date_features = create_date_features

def apply_on_series(df, func):
    return df.progress_apply(func)

def multicore_apply_by_chunks(df_st, func, max_chunks, max_workers = None, log_progress=False):
    max_workers = max_chunks if not max_workers else max_workers
    len_df = df_st.shape[0]
    n = len_df // max_chunks +1
    chunks = [df_st[i:i+n] for i in range(0,len_df,n)]
    fn_specs = []
    for i in chunks:
        fn_specs.append((apply_on_series, {"df" : i, 'func': func}))
    
    results, exceptions = run_in_parallel(fn_specs, threaded = False, max_workers = max_workers, log_at = 1, log_func = log_info, log_progress = log_progress)
    if len(exceptions) > 0 : print(exceptions[0][2])
    results = pd.concat(i[1] for i in results)
    del chunks
    del fn_specs
    del exceptions
    gc.collect()
    return results

pd.Series.multicore_apply_by_chunks = multicore_apply_by_chunks


from krezi.common import *
fit_to_screen() # when reload is called on krezi, this will restore the cell length

from krezi.logging_util.file_logger import Logger
from krezi.mssql_util.pymssql_dao import mssql_dao
from krezi.posgresql_util.posgresql_util import pgdao
from krezi.multiprocessing_util.mp_class_util import ParallelExecutor
from krezi.stopwatch_util.stopwatch import Stopwatch



"""

if linux : write this in .bashrc
if mac : write this in .zshrc

cmd : PYTHONPATH="$PYTHONPATH:/absolute/location/of/directory/krezi/"
cmd : echo PYTHONPATH

Module Funcs can be imported as 

1. import krezi --> runs __init__.py 
    common will be available as krezi.fit_to_screen() for ex
2. from krezi import * 
    common will be available as fit_to_screen() for ex
3. form krezi.common import *
4. import krezi.common as cf 
5. from krezi import Stopwatch

"""


from docx import Document
from datetime import datetime
from docx import Document
import argparse

from docx.enum.text import *
from docx.shared import Pt, Cm

date_suffix_dict = {}
date_suffix_dict.update({i:'st' for i in [1, 21, 31]})
date_suffix_dict.update({i:'nd' for i in [2, 12, 22]})

parser = argparse.ArgumentParser(
                    prog='Cover Letter Generator',
                    description='With a few variables, generates a pre worded cover letter')

parser.add_argument("-c", "--company", required = True,
                    help = "company you are applying for")
parser.add_argument("-r", "--role-name", required = True, 
                    help = "role for which you are applying for")
parser.add_argument("-s", "--source",  required = False,
                    default = 'LinkedIn', 
                    help = "where you found this role : Ex LinkedIn")
parser.add_argument("-i", "--is-internship", 
                    action='store_true', 
                    help = "is the role an internship")

args = parser.parse_args()

DATE_TODAY = datetime.now()

ROLE_NAME = args.role_name # role r

IS_INTERNSHIP = args.is_internship # internship i

COMPANY_NAME = args.company # company c

SOURCE = args.source # source s

MARGIN = 1.27 # Taken from Word : Narrow Margin 
LINE_SPACING = 1.15 # Taken from Word

if __name__ == '__main__':
    d = Document()

    header_line_1 = "Your Name"
    header_line_2 = "location | Phone | LinkedIn Url | Github Url"

    header = d.sections[0].header

    head_p0 = header.paragraphs[0]
    head_p0.text = header_line_1
    pf = head_p0.paragraph_format
    pf.alignment = WD_ALIGN_PARAGRAPH.CENTER
    head_p0.runs[0].bold = True
    fhp0 = head_p0.runs[0].font
    fhp0.size = Pt(18)
    fhp0.name = "Times New Roman"

    head_p1 = header.add_paragraph()
    head_p1.text = header_line_2
    pf = head_p1.paragraph_format
    pf.alignment = WD_ALIGN_PARAGRAPH.CENTER
    head_p1.runs[0].bold = False
    fhp1 = head_p1.runs[0].font
    fhp1.size = Pt(9)
    fhp1.name = "Times New Roman"

    paragraph_1 = "Subject"
    paragraph_2 = "Dear Manager,"
    paragraph_3 = "p3_text"
    paragraph_4 = "p4_text"
    paragraph_5 = "p5_text"
    paragraph_6 = "p6_text"
    paragraph_7 = "p6_text"
    paragraph_8 = "p7_text"
    ending_message = "Sincerely,\nYour Name"

    paragraph_1 = paragraph_1.replace(":role_name", ROLE_NAME)
    paragraph_3 = paragraph_3.replace(":role_name", ROLE_NAME)
    paragraph_3 = paragraph_3.replace(":source", SOURCE if SOURCE else "LinkedIn")
    paragraph_3 = paragraph_3.replace(":company_name", COMPANY_NAME)
    paragraph_7 = paragraph_7.replace(":company_name", COMPANY_NAME)

    p1 = d.add_paragraph(paragraph_1)
    fbp1 = p1.paragraph_format
    fbp1.alignment = WD_ALIGN_PARAGRAPH.LEFT
    fbp1.space_before = Pt(10)
    fbp1.space_after = Pt(10)
    fbp1.line_spacing = 1.15
    fbp1 = p1.runs[0].font
    fbp1.size = Pt(10)
    fbp1.name = "Times New Roman"

    p2 = d.add_paragraph(paragraph_2)
    fbp2 = p2.paragraph_format
    fbp2.alignment = WD_ALIGN_PARAGRAPH.LEFT
    fbp2.space_after = Pt(10)
    fbp2.line_spacing = 1.15
    fbp2 = p2.runs[0].font
    fbp2.size = Pt(10)
    fbp2.name = "Times New Roman"

    p3 = d.add_paragraph(paragraph_3)
    fbp3 = p3.paragraph_format
    fbp3.alignment = WD_ALIGN_PARAGRAPH.JUSTIFY
    fbp3.space_after = Pt(10)
    fbp3.line_spacing = 1.15
    fbp3 = p3.runs[0].font
    fbp3.size = Pt(10)
    fbp3.name = "Times New Roman"

    p4 = d.add_paragraph(paragraph_4)
    fbp4 = p4.paragraph_format
    fbp4.alignment = WD_ALIGN_PARAGRAPH.JUSTIFY
    fbp4.space_after = Pt(10)
    fbp4.line_spacing = 1.15
    fbp4 = p4.runs[0].font
    fbp4.size = Pt(10)
    fbp4.name = "Times New Roman"

    p5 = d.add_paragraph(paragraph_5)
    fbp5 = p5.paragraph_format
    fbp5.alignment = WD_ALIGN_PARAGRAPH.JUSTIFY
    fbp5.space_after = Pt(10)
    fbp5.line_spacing = 1.15
    fbp5 = p5.runs[0].font
    fbp5.size = Pt(10)
    fbp5.name = "Times New Roman"

    p6 = d.add_paragraph(paragraph_6)
    fbp6 = p6.paragraph_format
    fbp6.alignment = WD_ALIGN_PARAGRAPH.JUSTIFY
    fbp6.space_after = Pt(10)
    fbp6.line_spacing = 1.15
    fbp6 = p6.runs[0].font
    fbp6.size = Pt(10)
    fbp6.name = "Times New Roman"

    p7 = d.add_paragraph(paragraph_7)
    fbp7 = p7.paragraph_format
    fbp7.alignment = WD_ALIGN_PARAGRAPH.JUSTIFY
    fbp7.space_after = Pt(10)
    fbp7.line_spacing = 1.15
    fbp7 = p7.runs[0].font
    fbp7.size = Pt(10)
    fbp7.name = "Times New Roman"

    p8 = d.add_paragraph(paragraph_8)
    fbp8 = p8.paragraph_format
    fbp8.alignment = WD_ALIGN_PARAGRAPH.JUSTIFY
    fbp8.space_after = Pt(10)
    fbp8.line_spacing = 1.15
    fbp8 = p8.runs[0].font
    fbp8.size = Pt(10)
    fbp8.name = "Times New Roman"

    _ending_message = d.add_paragraph(ending_message)
    fb_ending_message = _ending_message.paragraph_format
    fb_ending_message.alignment = WD_ALIGN_PARAGRAPH.LEFT
    fb_ending_message.space_after = Pt(10)
    fb_ending_message.line_spacing = 1.15
    fb_ending_message = _ending_message.runs[0].font
    fb_ending_message.size = Pt(10)
    fb_ending_message.name = "Times New Roman"


    sections = d.sections
    for section in sections:
        section.top_margin = Cm(MARGIN)
        section.bottom_margin = Cm(MARGIN)
        section.left_margin = Cm(MARGIN)
        section.right_margin = Cm(MARGIN)

    d.save(f"/location_for_saving_file/Cover Letter {COMPANY_NAME} {ROLE_NAME} {DATE_TODAY:%Y-%m-%d}.docx")



import time
import pandas as pd

class Stopwatch:
    """
    A stopwatch class for recording time elapsed between events.

    This class allows you to record the time elapsed between consecutive calls
    to the `cycle` method. It also keeps track of the cumulative time elapsed
    since the initialization of the stopwatch. Event names can be optionally
    provided during each cycle to label the events in the log.

    Attributes:
        start_time (float or None): The time of the previous cycle call, or None for the first call.
        init_time (float or None): The time when the stopwatch was initialized, or None before the first cycle.
        event_log (list): A list of tuples containing (event_name, elapsed_time, cumulative_time) for each cycle.
        cumulative_time (float): The total time elapsed since the initialization of the stopwatch.

    Methods:
        cycle(event_name=None)
            Records the time elapsed since the previous call and updates the event log.
        reset()
            Resets the stopwatch to its initial state, clearing the event log.
        print_log()
            Prints the event log with elapsed times and cumulative times.
        get_log_dataframe()
            Returns a pandas DataFrame containing the event log.
    """

    def __init__(self):
        self.start_time = None # this changes every cycle
        self.init_time = None # this does not change ever unless reset is called
        self.event_log = []
        self.cumulative_time = 0

    def cycle(self, event_name=None):
        """
        Records the time elapsed since the previous call and updates the event log.

        Args:
            event_name (str, optional): The name of the event to be recorded. If not provided, the event will be recorded as "Unnamed Event".

        Returns:
            float: The time elapsed since the previous call.
        """
        current_time = time.time()
        if self.start_time is None:
            elapsed_time = 0
        else:
            elapsed_time = current_time - self.start_time

        if self.init_time is None:
            self.init_time = current_time

        self.cumulative_time += elapsed_time
        event_entry = (event_name, elapsed_time, self.cumulative_time)
        self.event_log.append(event_entry)

        self.start_time = current_time
        return elapsed_time

    def reset(self):
        """
        Resets the stopwatch to its initial state, clearing the event log.
        """
        self.start_time = None
        self.init_time = None
        self.event_log = []
        self.cumulative_time = 0

    def print_log(self):
        """
        Prints the event log with elapsed times and cumulative times.
        """
        if not self.event_log:
            print("No events recorded.")
        else:
            for event, elapsed_time, cumulative_time in self.event_log:
                event_name = event if event else "Unnamed Event"
                print(f"{event_name}: {elapsed_time:.6f} seconds (Cumulative: {cumulative_time:.6f} seconds)")

            total_time = time.time() - self.init_time
            print(f"Total time since initialization: {total_time:.6f} seconds")

    def get_log_dataframe(self):
        """
        Returns a pandas DataFrame containing the event log.

        Returns:
            pandas.DataFrame: A DataFrame with columns for the event name, elapsed time, and cumulative time.
        """
        data = [{"Event": event if event else "Unnamed Event",
                 "Elapsed Time (s)": elapsed_time,
                 "Cumulative Time (s)": cumulative_time}
                for event, elapsed_time, cumulative_time in self.event_log]
        return pd.DataFrame(data)


import concurrent.futures
import traceback

"""
ThreadPoolExecutor/ProcessPoolExecutor --> Executor Object:
    fn submit
    fn map
    fn shutdown

Executor.submit --> Future objects:
    fn cancel
    fn cancelled
    fn running
    fn done
    fn result
    fn exception
    fn add_done_callback
    fn set_running_or_notify_cancel
    fn set_result
    fn set_exception
"""

threaded = True

# common args
max_workers = None

# only thread pool
thread_name_prefix = ''

# only process pool
mp_context = None
max_tasks_per_child = None

# for logging
log_func = print
log_at = 10_000

def run_in_parallel(fn_specs, threaded=threaded, max_workers=max_workers, thread_name_prefix=thread_name_prefix, mp_context=mp_context, max_tasks_per_child=max_tasks_per_child, log_at=log_at, log_func=log_func, log_progress = False):
    """
    Run function concurrently or paralelly

    Process Pool works fine on Linux. For Mac, import function you want to process from a separate module.

    Args:
        fn_specs : list : required
            List of Tuples where each tuple holds function in 1st place and arguments for function as a dictionary in second place.
        threaded : bool : optional : default - True
            If you want to run the function concurrently, set this to True. This will set the executor to ThreadPoolExecutor.
            If you want to spawn a separate process for each function set this to False. This will set the executor to ProcessPoolExecutor.
        max_workers: int : optional 
            Works for ThreadPoolExecutor and ProcessPoolExecutor differently. Refer Concurrent.Futures documentation for more details.
        thread_name_prefix : optional : default - ''
            Keyword Argument for ThreadPoolExecutor. Refer Concurrent.Futures documentation for more details.
        mp_context: optional : optional : default - None
            Keyword Argument for ProcessPoolExecutor. Refer Concurrent.Futures documentation for more details.
        max_tasks_per_child : optional : default - None
            Keyword Argument for ProcessPoolExecutor. Refer Concurrent.Futures documentation for more details.
    Returns:
        results : list
            list of tuples where 1st element of tuple will be argument passed to function and 2nd element will be output from the function
        exceptions : list
            list of tuples where 1st element of tuple will be argument passed to function, 2nd element will be Exception Name, 
            and 3rd element will be traceback

    """
    # executor_fn = concurrent.futures.ThreadPoolExecutor if threaded else concurrent.futures.ProcessPoolExecutor

    func_list = [i[0] for i in fn_specs]
    kwargs_list = [i[1] for i in fn_specs]
    len_func_list = len(func_list)

    if threaded:
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers, thread_name_prefix=thread_name_prefix) as executor:
            futures = {}
            results = []
            exceptions = []
            for fn, kwargs in zip(func_list,kwargs_list):
                futures.update({executor.submit(fn, **kwargs) : kwargs})
            for num_future, future in enumerate(concurrent.futures.as_completed(futures)):
                if (log_progress) & (num_future%log_at == 0):
                    log_func(f"Execution of {num_future}/{len_func_list} complete.")
                try:
                    result = future.result()
                    results.append((futures[future], result))
                except Exception as e:
                    exc = type(e).__name__
                    exceptions.append((futures[future],exc,traceback.format_exc()))
        return results, exceptions

    if not threaded:
        with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp_context) as executor:
            futures = {}
            results = []
            exceptions = []
            for fn, kwargs in zip(func_list,kwargs_list):
                futures.update({executor.submit(fn, **kwargs) : kwargs})
            for num_future, future in enumerate(concurrent.futures.as_completed(futures)):
                if (log_progress) & (num_future%log_at == 0):
                    log_func(f"Execution of {num_future}/{len_func_list} complete.")
                try:
                    result = future.result()
                    results.append((futures[future], result))
                except Exception as e:
                    exc = type(e).__name__
                    exceptions.append((futures[future],exc,traceback.format_exc()))
        return results, exceptions


import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email import encoders
import sys
import traceback

class EmailNotifier:
    def __init__(self, smtp_server, smtp_port, sender_email, sender_password):
        """
        Initialize the EmailNotifier instance with SMTP server details and sender credentials.

        Args:
            smtp_server (str): SMTP server address. # smtp.gmail.com for gmail.
            smtp_port (int): SMTP server port. # For SSL, enter 465. For TLS, enter 587.
            sender_email (str): Sender's email address.
            sender_password (str): Sender's email password.
        """
        self.smtp_server = smtp_server
        self.smtp_port = smtp_port
        self.sender_email = sender_email
        self.sender_password = sender_password
        self.error_recipients = None

    def send_email(self, subject, body, recipients, cc=None, bcc=None, attachments=None):
        """
        Send an email with the provided subject, body, recipients, CC, BCC, and attachments.

        Args:
            subject (str): Email subject.
            body (str): Email body.
            recipients (str or list): Email recipients. Can be a single email or a list of emails.
            cc (str or list, optional): CC recipients. Can be a single email or a list of emails. Defaults to None.
            bcc (str or list, optional): BCC recipients. Can be a single email or a list of emails. Defaults to None.
            attachments (list, optional): List of file paths to attach. Defaults to None.
        """
        try:
            msg = MIMEMultipart()
            msg['From'] = self.sender_email
            msg['Subject'] = subject

            if isinstance(recipients, list):
                msg['To'] = ', '.join(recipients)
            else:
                msg['To'] = recipients

            if cc:
                if isinstance(cc, list):
                    msg['CC'] = ', '.join(cc)
                else:
                    msg['CC'] = cc

            if bcc:
                if isinstance(bcc, list):
                    msg['BCC'] = ', '.join(bcc)
                else:
                    msg['BCC'] = bcc

            msg.attach(MIMEText(body, 'plain'))

            if attachments:
                for attachment in attachments:
                    with open(attachment, 'rb') as file:
                        part = MIMEBase('application', 'octet-stream')
                        part.set_payload(file.read())
                        encoders.encode_base64(part)
                        part.add_header('Content-Disposition', f'attachment; filename={attachment}')
                        msg.attach(part)

            server = smtplib.SMTP(self.smtp_server, self.smtp_port)
            server.starttls()
            server.login(self.sender_email, self.sender_password)
            server.send_message(msg)
            server.quit()
            print("Email sent successfully!")
        except Exception as e:
            print(f"Error sending email: {str(e)}")
            self.send_error_email(str(e), traceback.format_exc())

    def send_error_email(self, error_message, traceback_info):
        """
        Send an error notification email with the provided error message and traceback information.

        Args:
            error_message (str): Error message.
            traceback_info (str): Traceback information.
        """
        subject = f"Script Failure Notification: {self.script_name}" if self.script_name else "Script Failure Notification"
        body = f"The script encountered an error:\n\n{error_message}\n\nTraceback:\n{traceback_info}"
        recipients = self.error_recipients or self.sender_email
        self.send_email(subject, body, recipients)

    def set_error_trigger(self, script_name, recipients=None):
        """
        Set the global exception handler to the `global_exception_handler` method of the EmailNotifier instance.

        Args:
            recipients (str or list, optional): Email recipients for error notifications.
                Can be a single email or a list of emails. Defaults to None.
        """
        self.script_name=script_name
        self.error_recipients = recipients
        sys.excepthook = self.global_exception_handler

    def global_exception_handler(self, exctype, value, tb):
        """
        Global exception handler that sends an error notification email and exits the script with a non-zero status code.

        Args:
            exctype: Exception type.
            value: Exception value.
            tb: Traceback object.
        """
        self.send_error_email(str(value), ''.join(traceback.format_exception(exctype, value, tb)))
        print(f"An error occurred: {str(value)}")
        print("Traceback:")
        traceback.print_exception(exctype, value, tb)
        sys.exit(1)


# from az_common_funcs import *

import pandas as pd
import psycopg2
from psycopg2.extras import RealDictCursor

class pgdao:
    def __init__(self, conn_dict, default_db = 'airbnb'):
        self.conn_dict = conn_dict
        self.database = default_db if conn_dict.get('database') is None else conn_dict.get('database')
        self.host = conn_dict.get('host')
        self.port = conn_dict.get('port')
        self.user = conn_dict.get('user')
        self.password = conn_dict.get('password')
        self.conn = None
        self.cursor = None
    
    def _init_connection(self):
        self.conn = psycopg2.connect(
                                host = self.host,
                                port = self.port,
                                database = self.database,
                                user = self.user,
                                password = self.password)   
        self.conn.set_session(readonly=True)  
        
        
    def _init_cursor(self):
        if not self.conn or self.conn.closed!=0:
            self._init_connection()
        self.cursor = self.conn.cursor(cursor_factory=RealDictCursor)
    
    def change_database(self,database):
        self.database = database
        self._init_connection()

    def _query_builder(self, query_str, params):
        if not isinstance(query_str, str): 
            raise TypeError(f"query should be of type str but {type(query_str).__name__} was passed!")
        if not isinstance(params, dict): 
            raise TypeError(f"params should be of type dict but {type(params).__name__} was passed!") 

        for k in params:
            query_str = query_str.replace(f":{k}", f"%({k})s")
        
        return query_str
    
    def query(self, query_str, params = {}, dry_run=False, show_query=True):
        final_query = self._query_builder(query_str, params)
        if not self.cursor or self.cursor.closed != 0:
            self._init_cursor()
        self.conn.rollback()
        if dry_run:
            print(f"Dry Run Query : \n{self.cursor.mogrify(final_query, params)!r}")
        else:
            if show_query:
                print(f"Executing Query : \n{self.cursor.mogrify(final_query, params)!r}")
            self.cursor.execute(final_query, params)
            rows = self.cursor.fetchall()
            res_df = pd.DataFrame(rows)
            res_df.columns = res_df.columns.str.upper()
            self.conn.rollback()
            return res_df
    
    def query_fetch_in_batches(self, query_str, params = {}, dry_run=False, batch_size=100_000, max_size=500_000):
        final_query = self._query_builder(query_str, params)
        if not self.cursor or self.cursor.closed != 0:
            self._init_cursor()
        self.conn.rollback()
        if dry_run:
            print(f"Dry Run Query : \n{self.cursor.mogrify(final_query, params)!r}")
        else:
            print(f"Executing Query : \n{self.cursor.mogrify(final_query, params)!r}")
            self.cursor.execute(final_query, params)
            df_list = []
            row_count = min(self.cursor.rowcount, max_size)
            print(f"Rows found :: {self.cursor.rowcount}, Rows to retrieve :: {row_count}")
            while self.cursor.rownumber < row_count:
                fetch_size = min(batch_size, row_count-self.cursor.rownumber)
                print(f"Fetching next {fetch_size:_} rows")
                df_list.append(pd.DataFrame(self.cursor.fetchmany(fetch_size)))
            self.conn.rollback()
            return pd.concat(df_list, ignore_index=True)
    

            


import pandas as pd
import sqlalchemy

class mssql_dao:
    """
    Class to run Queries on MSSQL DB Schema Provided
    """
    def __init__(self, conn_dict, schema = 'vbay'):
        self.conn_dict = conn_dict 
        self.schema = schema 
        self.engine = None
        self.cursor = None
    
    def init_engine(self):
        cnxn_str = f"mssql+pymssql://{self.conn_dict['user']}:{self.conn_dict['password']}@{self.conn_dict['server']}/{self.schema}"
        print(f"Trying to establish a connection with connection string \n{cnxn_str}")
        self.engine = sqlalchemy.create_engine(cnxn_str)
        
    def init_cursor(self):
        if not self.engine:
            self.init_engine()
        print("Initializing Cursor")
        self.cursor = self.engine.raw_connection().cursor(as_dict=True)
    
    def change_schema(self,schema):
        print(f"Changing Schema from {self.schema} to {schema}")
        self.schema = schema
        self.init_engine()
        self.init_cursor()
        
    @staticmethod
    def query_builder(query_str, params):
        """
        A simple function that returns the final query that can be executed with actually executing the query
        Return: String

        Note: This is a static method. It is safe to use from outside the class
        """
        if not isinstance(query_str, str): 
            raise TypeError(f"query should be of type str but {type(query_str).__name__} was passed!")
        if not isinstance(params, dict): 
            raise TypeError(f"params should be of type dict but {type(params).__name__} was passed!") 

        for k in params:
            query_str = query_str.replace(f":{k}", f"%({k})s")
        
        return query_str
    
    def query(self, query_str, params = {}):
        """
        Inputs:     
            query_str : str : query as is written in sql client
            params : dict : default empty : pass parameters in a dictionary and refer to them in query_str using a ':'
        Returns:
            DataFrame
        Note: As a good practice, only call this function from outside
        """
        final_query = self.query_builder(query_str, params)
        if not self.cursor:
            self.init_cursor()
        if not self.cursor.connection:
            self.init_cursor()
        self.cursor.execute(final_query, params)
        rows = self.cursor.fetchall()
        res_df = pd.DataFrame(rows)
        res_df.columns = res_df.columns.str.replace(" ","_").str.upper() # uppercasing all columns as well as replacing any whitespaces with underscores
        return res_df


import logging
import sys

class Logger:
    def __init__(self, logger_name=None, add_stream_handler = True, std_err = False, level = 'debug', **kwargs):
        self.logger_name = logger_name if logger_name else __name__
        self.add_stream_handler = add_stream_handler
        self.std_err = std_err
        self.level = self.LEVELS.get(level, logging.INFO)
        self.logger = None
        self.dt_fmt_basic = kwargs.get('dt_fmt_basic', True)
    
    LEVELS = {'debug': logging.DEBUG, 'info': logging.INFO, 'warning': logging.WARNING, 'error': logging.ERROR, 'critical': logging.CRITICAL}
    
    @staticmethod
    def remove_handlers(logger):
        while logger.handlers:
            handler = logger.handlers[0]
            handler.close()
            logger.removeHandler(handler)
        return logger
    
    def set_logger(self, dt_fmt_basic=None):
        logging.basicConfig(
            format="%(asctime)s :: [%(levelname)s] :: %(message)s", 
            datefmt='%d %B, %Y %I:%M:%S %p %z',
            level=logging.ERROR,
            stream = sys.stderr,
            force=True)
        logging.getLogger().removeHandler(logging.getLogger().handlers[0])
        self.logger = logging.getLogger(self.logger_name) ## IMP: __name__ is important for scope of logger
        self.logger.setLevel(logging.DEBUG)
        dt_fmt_basic = self.dt_fmt_basic if dt_fmt_basic is None else dt_fmt_basic
        if dt_fmt_basic:
            formatter = logging.Formatter("%(asctime)s :: [%(levelname)s] :: %(message)s")
        else:
            formatter = logging.Formatter("%(asctime)s :: [%(levelname)s] :: %(message)s", datefmt='%d %B, %Y %I:%M:%S %p %z')
        if self.add_stream_handler:
            if not self.std_err:
                # self.logger = self.remove_handlers(self.logger)
                stream_handler = logging.StreamHandler(stream = sys.stdout)
            else:
                stream_handler = logging.StreamHandler(stream = sys.stderr)
            stream_handler.setFormatter(formatter)
            stream_handler.setLevel(self.level)
            self.logger.addHandler(stream_handler)
    
    def get_logger(self, dt_fmt_basic = True):
        if not self.logger:
            self.set_logger(dt_fmt_basic)
        return self.logger
    
    def add_file_handler(self, filepath, level = logging.INFO, dt_fmt_basic=None):
        file_handler = logging.FileHandler(filepath)
        file_handler.setLevel(level)
        dt_fmt_basic = self.dt_fmt_basic if dt_fmt_basic is None else dt_fmt_basic
        if dt_fmt_basic:
            formatter = logging.Formatter(
                '%(asctime)s :: %(name)s :: %(levelname)s :: %(message)s')
        else:
            formatter = logging.Formatter(
                '%(asctime)s :: %(name)s :: %(levelname)s :: %(message)s', datefmt='%d %B, %Y %I:%M:%S %p %z')
        file_handler.setFormatter(formatter)
        self.logger.addHandler(file_handler)
        return self.get_logger()
    
    def add_log_prefix(self, prefix):
        self.prefix = prefix
        for handler in self.logger.handlers:
            if self.prefix:
                formatter = logging.Formatter(handler.formatter._fmt.replace("%(message)s", "") + f"{prefix} :: %(message)s", datefmt=handler.formatter.datefmt)
                handler.setFormatter(formatter)
        return self.get_logger()
    
    def remove_log_prefix(self):
        for handler in self.logger.handlers:
            if self.prefix:
                formatter = logging.Formatter(handler.formatter._fmt.replace(f"{self.prefix} :: %(message)s", "%(message)s"), datefmt=handler.formatter.datefmt)
                handler.setFormatter(formatter)
        self.prefix = None
        return self.get_logger()
    
    def __str__(self):
        to_print("logger not initialized")
        if self.logger:
            to_print = ""
            to_print = "".join([to_print,f"Logger is set. Name :: {self.logger.name} with the following handlers :"])
            for num, i in enumerate(self.logger.handlers, start = 1):
                h_name, h_op, h_level = i.__str__().replace("<","").replace(">","").split(" ")
                to_print = "".join([to_print, "\n", f"{num} :: {h_name} :: {h_level}"])
        return to_print


import concurrent.futures
import traceback

"""
ThreadPoolExecutor/ProcessPoolExecutor --> Executor Object:
    fn submit
    fn map
    fn shutdown

Executor.submit --> Future objects:
    fn cancel
    fn cancelled
    fn running
    fn done
    fn result
    fn exception
    fn add_done_callback
    fn set_running_or_notify_cancel
    fn set_result
    fn set_exception
"""

class ParallelExecutor:
    def __init__(self, threaded=True, max_workers=None, thread_name_prefix='', mp_context=None, max_tasks_per_child=None, log_func=print, log_at=10_000):
        self.threaded = threaded
        self.max_workers = max_workers
        self.thread_name_prefix = thread_name_prefix
        self.mp_context = mp_context
        self.max_tasks_per_child = max_tasks_per_child
        self.log_func = log_func
        self.log_at = log_at

    def run_in_parallel(self, fn_specs, log_progress=False):
        """
        Run function concurrently or paralelly

        Process Pool works fine on Linux. For Mac, import function you want to process from a separate module.

        Args:
            fn_specs : list : required
                List of Tuples where each tuple holds function in 1st place and arguments for function as a dictionary in second place.
            log_progress : bool : optional : default - False
                If True, it will log the progress of execution at intervals specified by log_at.
        Returns:
            results : list
                list of tuples where 1st element of tuple will be argument passed to function and 2nd element will be output from the function
            exceptions : list
                list of tuples where 1st element of tuple will be argument passed to function, 2nd element will be Exception Name,
                and 3rd element will be traceback
        """
        func_list = [i[0] for i in fn_specs]
        kwargs_list = [i[1] for i in fn_specs]
        len_func_list = len(func_list)

        if self.threaded:
            with concurrent.futures.ThreadPoolExecutor(max_workers=self.max_workers, thread_name_prefix=self.thread_name_prefix) as executor:
                futures = {executor.submit(fn, **kwargs): kwargs for fn, kwargs in zip(func_list, kwargs_list)}
                return self._process_futures(futures, len_func_list, log_progress)

        if not self.threaded:
            with concurrent.futures.ProcessPoolExecutor(max_workers=self.max_workers, mp_context=self.mp_context) as executor:
                futures = {executor.submit(fn, **kwargs): kwargs for fn, kwargs in zip(func_list, kwargs_list)}
                return self._process_futures(futures, len_func_list, log_progress)

    def _process_futures(self, futures, len_func_list, log_progress):
        results = []
        exceptions = []
        for num_future, future in enumerate(concurrent.futures.as_completed(futures)):
            if log_progress and num_future % self.log_at == 0:
                self.log_func(f"Execution of {num_future}/{len_func_list} complete.")
            try:
                result = future.result()
                results.append((futures[future], result))
            except Exception as e:
                exc = type(e).__name__
                exceptions.append((futures[future], exc, traceback.format_exc()))
        return results, exceptions


from setuptools import setup, find_packages

with open("README.md", 'r') as f:
    description = f.read()

setup(
    name='krezi',
    version='1.2',
    python_requires='>= 3.9',
    author="Aadil Zikre",
    author_email="aadilzikre@gmail.com",
    packages=find_packages(),
    install_requires=[
        'numpy>=1.23.0',
        'pandas>=1.4.2',
        'itables>=1.5.2',
        'tqdm'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Programming Language :: Python :: 3'
    ],
    description="Array of Utilities for Lazy but Efficient Programmers",
    long_description=description,
    long_description_content_type='text/markdown'
)

## To Build the Project : in the directory, run python setup.py sdist bdist_wheel
## Remove the previous build manually from dist/
## Upload the Project on PyPI : 